package umdes;


import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JToolBar;
import javax.swing.ListSelectionModel;
import javax.swing.TransferHandler;



public class EventList extends JDialog implements ActionListener {
	
	private static final long serialVersionUID = 6732762092107736146L;
	
	private JList list;
    private Vector internalList; //for modifications to the list
    private Vector eventOrder;
    private boolean orderStates;
    private boolean curOrderStates = false;
    private boolean curOrderEvents = false;
    private final int EVENTLIST_WIDTH = 200;
    private final int EVENTLIST_HEIGHT = 400;
    private JToolBar helpToolBar;
    private String delimString = "@@@@";
    private JButton helpButton;
    private JPanel mainPanel;
    private boolean sortString = false;
    
    private String help_events_only = "Arrange the events so that the ones at the top are " +
    			"considered first\n\nAn event or multiple events can be selected and "+
    			"then dragged to the desired\nposition.  Also, \"Order by Events\" will place " + 
    			"the events in alphanumeric order.\nClicking it again will reverse this "+ 
    			"order.";
    
    private String help_with_states = "Arrange the events so that the ones at the top are " +
				"considered first.\nThe first column contains states while the second contains events.\n\n" +
				"An event or multiple events can be selected and "+
				"then dragged to\nthe desired position.  Also, \"Order by Events\" will place " + 
				"the events in\n alphanumeric order.  Clicking it again will reverse this "+ 
				"order.\n\"Order by States\" puts the states in alphanumeric order.";
    
    
    public EventList(Vector<?> s, JFrame parent, boolean orderAllEvents) {
    	
    	super(parent, "Events" , true); 

    	internalList = new Vector();
    	orderStates = orderAllEvents;
 
 
    	if(s.get(0) instanceof String) sortString = true;
        
        final JButton acceptButton = new JButton("Accept");
        acceptButton.setActionCommand("Accept");
        acceptButton.addActionListener(this);

    	final JButton stateOrderButton = new JButton("Order by State");
        if(orderAllEvents){
        	stateOrderButton.setActionCommand("stateOrder");
        	stateOrderButton.addActionListener(this);
        }
        
        final JButton eventOrderButton = new JButton("Order by Event");
        eventOrderButton.setActionCommand("eventOrder");
        eventOrderButton.addActionListener(this);
    	
        mainPanel = new JPanel();
        mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.PAGE_AXIS));
        mainPanel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
    	
    	helpButton = new JButton("Help");
    	helpButton.setActionCommand("Help");
    	helpButton.addActionListener(this);   
   
        helpToolBar = new JToolBar(); 
        helpToolBar.setFloatable(false); 
        helpToolBar.add(Box.createHorizontalGlue());
        helpToolBar.add(helpButton);
        mainPanel.add(helpToolBar);
    
    	list = new JList(s);
    	list.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        list.setLayoutOrientation(JList.VERTICAL);
        list.setVisibleRowCount(-1);
        JScrollPane listScroller = new JScrollPane(list);
        listScroller.setPreferredSize(new Dimension(EVENTLIST_WIDTH, EVENTLIST_HEIGHT));
        listScroller.setAlignmentX(LEFT_ALIGNMENT);
        
        for(int i = 0; i<list.getModel().getSize(); i++){
        	internalList.add(list.getModel().getElementAt(i));
        }
        
        list.setTransferHandler(new TransferEvent()); //a custom private TransferHandler defined below
        
        list.setDragEnabled(true);
        
        
        JScrollPane listView = new JScrollPane(list);
        listView.setPreferredSize(new Dimension(EVENTLIST_WIDTH, EVENTLIST_HEIGHT));
        
        JPanel panel = new JPanel(new BorderLayout());
        panel.add(listView, BorderLayout.CENTER);
       
        if(orderStates) panel.setBorder(BorderFactory.createTitledBorder("Place Transitions in Desired Order"));   
        else panel.setBorder(BorderFactory.createTitledBorder("Place Events in Desired Order"));  
        mainPanel.add(panel);
        
        
        //Create and set up the window.
        setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
        setResizable(false);

        //Lay out the buttons from left to right.
        JPanel buttonPane = new JPanel();
        buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
        if(orderAllEvents) buttonPane.add(stateOrderButton);
        buttonPane.add(eventOrderButton);
 
        
        JPanel upAndDown = new JPanel();
        upAndDown.setLayout(new BoxLayout(upAndDown, BoxLayout.X_AXIS));
        final JButton UpButton = new JButton("Up");
        UpButton.setActionCommand("Up");
        UpButton.addActionListener(this);  
        final JButton DownButton = new JButton("Down");
        DownButton.setActionCommand("Down");
        DownButton.addActionListener(this);
        upAndDown.add(acceptButton);
        upAndDown.add(UpButton);
        upAndDown.add(DownButton);
        
        mainPanel.add(upAndDown);       
        mainPanel.add(buttonPane);


        mainPanel.setOpaque(true); //content panes must be opaque
        setContentPane(mainPanel);

        //Display the window.
        pack();
        centerOnScreen();
        setVisible(true);          	
    }
  
    /**
     * centers the ObserveList on the Screen
     */
    private void centerOnScreen(){
        Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
        
        // Determine the new location of the window
        int w = getSize().width;
        int h = getSize().height;
        int x = (dim.width-w)/2;
        int y = (dim.height-h)/2;
        
        // Move the window
        setLocation(x, y);
    }
    
    /**
     * handles all of the buttons that can be pressed
     */
    public void actionPerformed(ActionEvent e) {
        if ("Accept".equals(e.getActionCommand())) {
        	setEventOrder(new Vector());
        	for(int i = 0; i < internalList.size(); i++){
            	//eventOrder.add((String) internalList.get(i));
        		getEventOrder().add(internalList.get(i));     	
        	}  
            setVisible(false);
        } else if("stateOrder".equals(e.getActionCommand())) {
        	if(!curOrderStates){
          	    if(internalList.get(0) instanceof String) Collections.sort(internalList);
        		if(internalList.get(0) instanceof Window){
        			Comparator c = Window.getComparatorState(false);
        			Collections.sort(internalList , c);
        		}
            	curOrderStates = true;
            	curOrderEvents = false;
        	} else {       		
        		if(internalList.get(0) instanceof String) Collections.sort(internalList, Collections.reverseOrder());
        		if(internalList.get(0) instanceof Window){
        			Comparator c = Window.getComparatorState(true);
        			Collections.sort(internalList , c);
        		}
            	curOrderStates = false;         	    
        	}
        	list.setListData(internalList);
        	
        } else if ("eventOrder".equals(e.getActionCommand())) {
        	if(!curOrderEvents){
            	if(orderStates){
            		Comparator c = null;
            		if(sortString) c = new compareEvents();
            		else if(internalList.get(0) instanceof Window){
            			c = Window.getComparatorEvent(false);
            		}
            		if(c==null)return;
            		Collections.sort(internalList, c);

            	} else {
              	    Collections.sort(internalList);       		
            	}
            	curOrderEvents = true;
            	curOrderStates = false;
        	} else {
            	if(orderStates){
            		Comparator c = null;
            		if(sortString){
            			c = new compareEvents();
                		((compareEvents) c).reverse = true;
            		}
            		else if(internalList.get(0) instanceof Window){
            			c = Window.getComparatorEvent(true);
            		}
            		if(c==null)return;
            		Collections.sort(internalList, c);

            	} else {
            		if(internalList.get(0) instanceof String) Collections.sort(internalList , Collections.reverseOrder());
            		if(internalList.get(0) instanceof Window){
            			Comparator c = Window.getComparatorState(true);
            			Collections.sort(internalList , c);
            		}
            	}  
            	curOrderEvents = false;
        	}
    		list.setListData(internalList);       	
        } else if ("Up".equals(e.getActionCommand())) {
        	int top = 0;
        	int[] indices = list.getSelectedIndices();
        	if(indices.length == 0|| indices[0] == -1) return;
        	for(int i = 0; i < indices.length; i++){
        		if(indices[i] > top){
        			Object o = internalList.remove(indices[i]);
        			indices[i] = indices[i] - 1;
        			internalList.add(indices[i] , o); 
        		} else {top++;}
        	}
    		list.setListData(internalList); 
    		list.setSelectedIndices(indices);
        	
        } else if ("Down".equals(e.getActionCommand())) {
        	int bottom = internalList.size() - 1;
        	int[] indices = list.getSelectedIndices();
        	if(indices.length == 0|| indices[0] == -1) return;
        	for(int i = indices.length - 1; i >= 0; i--){
        		if(indices[i] <  bottom){
        			Object o = internalList.remove(indices[i]);
        			indices[i] = indices[i] + 1;
        			internalList.add(indices[i] , o); 
        		} else {bottom--;}
        	}
    		list.setListData(internalList); 
    		list.setSelectedIndices(indices);
        } else if ("Help".equals(e.getActionCommand())) {
        	if(orderStates){
        		JOptionPane.showMessageDialog(this, help_with_states,
        				"Order Events", JOptionPane.PLAIN_MESSAGE);
        	} else { 
        		JOptionPane.showMessageDialog(this, help_events_only,
        				"Order Events", JOptionPane.PLAIN_MESSAGE);
        	}
        }
        
    }
   
    public void setEventOrder(Vector eventOrder) {
		this.eventOrder = eventOrder;
	}

	public Vector getEventOrder() {
		return eventOrder;
	}

	private class TransferEvent extends TransferHandler{

		private static final long serialVersionUID = 1L;

		public int getSourceActions(JComponent c) {
            return MOVE;
        }
        
        protected Transferable createTransferable(JComponent c) {
            JList list = (JList)c;
            Object[] values = list.getSelectedValues();        
            StringBuffer buff = new StringBuffer();

            if(values.length==0) return null;
            if(values[0] instanceof Window){
            	DataFlavor flavor = new DataFlavor(Window.class, "Window");
            	flavor.setHumanPresentableName("Window");
            	Transferable t = null;
            	for(int i = 0; i < values.length; i++) Window.transferableWindows.add(values[i]);
            	try{
            		t = (Transferable) ((Window) values[0]).getTransferData(flavor);
            	}catch(Exception e){ return null;}
            	return t;
            }
            for (int i = 0; i < values.length; i++) {
                Object val = values[i];
                buff.append(val == null ? "" : val.toString());
                buff.append(delimString);
            }
            return new StringSelection(buff.toString());
        }
    	
        public boolean canImport(JComponent comp, DataFlavor[] transferFlavors){
        	return true;
        }
    	
        public boolean importData(JComponent comp, Transferable t) {

        	int indexTo = 0;
        	int removeFromTop = 0;
        	int removeFromBottom = 0;
    		Object o = null;
    		boolean flavor_is_string = true;
    		Object[] selectedObj = null;
        	try{
        		o = t.getTransferData(DataFlavor.stringFlavor);  
        		flavor_is_string = true;
        	} catch(Exception e){}
        	if(!(o instanceof String)){
            	try{
            		DataFlavor d = new DataFlavor(Window.class, "Window");
            		o = t.getTransferData(d); 
            		flavor_is_string = false;
            	} catch(Exception e){
            		return false;
            	}
        	}
        	
        	if(o==null) return false;
        	
        	if(flavor_is_string) selectedObj = ((String) o).split(delimString);
        	else{
        		selectedObj = new Object[Window.transferableWindows.size()];
        		for(int i = 0; i < Window.transferableWindows.size(); i++) selectedObj[i] = Window.transferableWindows.get(i);
        		Window.transferableWindows.clear();
        	}
        	if(comp instanceof JList){
        		indexTo = ((JList) comp).getSelectedIndex();
        		if(indexTo==-1)return false;
        	}
        	int[] indicesFrom = new int[selectedObj.length];
        	for(int i = 0; i < selectedObj.length; i++){
        		indicesFrom[i] = internalList.indexOf(selectedObj[i]);
        		if(indicesFrom[i] == -1) return false;
        	}
        	
        	
        	for(int i = 0; i< selectedObj.length; i++){
        		if(internalList.contains(selectedObj[i])){
        			if(indicesFrom[i] <= indexTo){
            			internalList.remove(internalList.indexOf(selectedObj[i]));
            			internalList.add(indexTo, selectedObj[i]); 
        				removeFromTop++;
        			}
        			else{
        				if(removeFromTop > 0 && removeFromBottom==0) removeFromBottom++;
            			internalList.remove(internalList.indexOf(selectedObj[i]));
            			internalList.add(indexTo + removeFromBottom, selectedObj[i]); 
        				removeFromBottom++;
        			}
        			

        		}            		
        	}
        	
        	((JList) comp).setListData(internalList);

        	curOrderStates = false;
        	curOrderEvents = false;
        	
        	int[] selected = new int[selectedObj.length];
        	if(removeFromTop > 0) removeFromTop--;
        	for(int i = 0; i < selected.length; i++){
        		selected[i] = indexTo + i - removeFromTop;
        	}
        	list.setSelectedIndices(selected);
        	
    		return true;
    	}
      	
       	
    }
    
    private class compareEvents implements Comparator{

    	public boolean reverse = false;
    	public int compare(Object obj1, Object obj2){
    		int index1 = ((String) obj1).trim().lastIndexOf(' ');
    		int index2 = ((String) obj2).trim().lastIndexOf(' ');
    		String str1 = ((String) obj1).trim().substring(index1);
    		String str2 = ((String) obj2).trim().substring(index2);
    		if(reverse==true){
    			return str1.compareTo(str2) * -1;
    		}
    		else return str1.compareTo(str2);
    	}
    }

}
