package umdes;

import giddes.xml.automaton.Automaton;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.Unmarshaller;

public class OnlineMinSenAct {

	/** The location where the JAXB auto-generated files are stored. These files help the XmlFile to save and load from the XML format. */
	private static final String UNMARSHAL_LOCATION = "giddes.xml.automaton";
	private static UmdesAutomaton automaton;
	private static String fileName;
	private static InputWindow inputWindow;
	private static BufferedWriter outOnlineMinSen;
	private static Vector<String> EventOrder; 
	private static Vector<RegExpr> transOrder; 
	private static String workDir;
	
	public static void main(String[] args) {

		if(args.length!=1){
			System.out.println("Error");
			return;
		}
		
		fileName = args[0];
		try{
			JAXBContext jc = JAXBContext.newInstance(UNMARSHAL_LOCATION);
	        Unmarshaller u = jc.createUnmarshaller();
	        Automaton xmlData = (Automaton) u.unmarshal(new FileInputStream(fileName));
			automaton = new UmdesAutomaton(xmlData);			
		} catch(Exception e){
			System.out.println("Error loading file");
			return;
		}
		File f = new File(fileName);
		workDir = f.getParent();
		if(workDir==null)workDir = System.getProperty("user.dir");
		
		EventOrder = new Vector<String>();
		transOrder = new Vector<RegExpr>();
    	Vector<RegExpr> inputString = new Vector<RegExpr>();
    	
		inputWindow = new InputWindow();
		inputWindow.setHeaderText("Enter Spec");
		inputWindow.setVisible(true);
		
		//waits for the submit button to be pressed
		while(!inputWindow.isSubmitPressed()) Thread.yield();
		
	  	try{
	  		automaton.enterSpec(formatSpec(inputWindow.getText()));
	  	}catch(Exception e){
	  		System.out.println("Invalid Spec");
	  		JOptionPane.showMessageDialog(null, e.getMessage(),
					  "Error:", JOptionPane.ERROR_MESSAGE);
	  		System.exit(0);
	  	}	
    	
	  	System.out.println("compute minimum sensor activity");
    	if(!automaton.getTSpec().isEmpty()){     	  

    		
    		Vector<RegExpr> events = new Vector<RegExpr>();
    		Qc local_stateSet = new Qc();
    			
    		boolean quitTest = false;
    		//initial call to min_sen_act
    		try {
				computeMinSenAct(null);
			} catch (IOException e1) { 
				System.out.println("failed compute");
				System.exit(1); 
				}
    			
    		boolean display_unobservable = true;
    		int display_answer = JOptionPane.showConfirmDialog(inputWindow,
    				"Should unobservable events and events with"+
    				" there sensors off be displayed?", "Display",
    				JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
    		if(display_answer == JOptionPane.NO_OPTION)
    			display_unobservable = false;
    		else
    			display_unobservable = true;
    			 		
    		while(!quitTest){
    			quitTest = true;  //if exits with cancel it will break loop
    				
    			local_stateSet = automaton.getStateSet();
    			events = automaton.getAlphabet();
    			
    			for(int i = 0; i< events.size(); i++){
    				if(!automaton.canTransition(local_stateSet, (RegExpr) events.get(i))){ events.remove(i); i--;}
    			}
    			//remove the events that are always unobservable
    			if(!display_unobservable){
    				for(int i = 0; i<events.size(); i++){
    					if(local_stateSet.E_uo.contains((RegExpr) events.get(i))){ events.remove(i); i--;}
    					if(i<0)continue;
    					if(local_stateSet.E_ua.contains((RegExpr) events.get(i))){ events.remove(i); i--;}
    				}
    			}
    			
    			//the plus one is for the quit string
    			String[] possibleValues = new String[events.size()+1];
    			for ( int i = 0; i < events.size(); i++){
    				if(local_stateSet.E_a.contains((RegExpr) events.get(i)))
    					possibleValues[i] = "(" + ((RegExpr) events.get(i)).getExpression() + ")";
    				if(display_unobservable&&local_stateSet.E_ua.contains((RegExpr) events.get(i)))
    					possibleValues[i] = "(" + ((RegExpr) events.get(i)).getExpression() + ")  (sensors off)";
    				if(display_unobservable&&local_stateSet.E_uo.contains((RegExpr) events.get(i)))
    					possibleValues[i] = "(" + ((RegExpr) events.get(i)).getExpression() + ")  (unobservable)"; 
   				}
    			possibleValues[events.size()] = "quit";
    			
    			// prompt the user for which failure type to use
    			String title = "Enter next event in input string";
    			String description = "current states --> {";
    				
    			for(int i =0; i<local_stateSet.states.size() - 1;i++){
    				description = description + local_stateSet.states.get(i).getName() + ", ";
    			}
    			description = description + 
    			local_stateSet.states.get(local_stateSet.states.size()-1).getName() + "}";
    			      
    			int answer = 0;
    			//if there is no partition then the program exits
    			if(events.size()>=0){
    				answer =  getUserInput(title, description, possibleValues, 
    						(possibleValues == null || possibleValues.length == 0) ? null : possibleValues[0]);
    			}
    			// if they cancel the answer will be 0
    			if ( answer == events.size() + 1||answer==0){ quitTest = true; continue;}
    			else quitTest = false;
    				

    			if(local_stateSet.E_a.contains(events.get(answer-1))){
    				try {
						computeMinSenAct(events.get(answer-1));
					} catch (IOException e1) { System.exit(1); }
					
    				inputString.add(events.get(answer-1));
    			}
    					     			
    		}
    		try {
				automaton.printInputString(inputString, outOnlineMinSen);
			} catch (IOException e) {System.exit(1);}
    		automaton.clearSpec();
    			
    			
    	}
    	inputString.clear();

    	inputWindow.dispose();
		System.exit(0);
	}

	    /**
	     * 
	     * This function will call the min_sen_act algorithm and output the results to online_min_sen_act.out
	     * @param workDir : the place where the file is 
	     * @param r : the observed event
	     * @throws IOException
	     */  
	    
	    public static void computeMinSenAct(RegExpr r) throws IOException{

	  	    //initialize the stateSet to start at the initial state
	      	if(r==null){
	      		System.out.println("create output file..." );
	      		System.out.println(workDir + System.getProperty("file.separator") +  "online_min_sen_act.out");
	      		File f = new File(workDir + System.getProperty("file.separator") +  "online_min_sen_act.out");
	      		outOnlineMinSen = new BufferedWriter(new FileWriter(f.getAbsolutePath()));     
	      		System.out.println("order transitions");
	  		    orderTransitions();
	  		    System.out.println("print to file");
	  		    printTSpec(outOnlineMinSen); //prints the T_spec to the file
	      		Vector<umdes.State> initial_state = new Vector<umdes.State>();
	      		initial_state.add(automaton.getInitial());
	      		getStateSet().states = automaton.unObservableReach(initial_state,null);
	  		    outOnlineMinSen.write("\nINITIAL\n");   
	  		    System.out.println("complete compute");
	      	}
	  	    	  
	      	//do the transition of event r
	  	    if(r!=null){
	  	    	getStateSet().clearEvents();  
	  	    	automaton.setStateSet(automaton.doTransition(getStateSet(),r));
	  		    outOnlineMinSen.write("\nEVENT OBSERVED: " + r.expression + "\n");
	  	    }
	  	    
		  
	      	outOnlineMinSen.write("States-------->  " + getStateSet().outputStates() + "\n");  	//output the states before
	  	    min_sen_act();
	  	  	outOnlineMinSen.write("Turned off:  " + getStateSet().policyEUA() + "\n");
	  	  	outOnlineMinSen.write("Kept on:     " + getStateSet().policyEA() + "\n");
	  	  	outOnlineMinSen.write("Sensors off--->  " + getStateSet().outputStates() + "\n");
	  	    
	  	    getStateSet().states = automaton.unObservableReach(getStateSet().states,null);

	    }
	    
	    /**
	     * changes the stateSet of this Fa to include the activation policy in its
	     * hashSets (E_a, E_ua, E_uo). It then changes the stateSet.states to include
	     * the added states by the turn off sensors from the activation policy.
	     * StateSet.states_unchanged remains the same as states before this function.
	     * 
	     */
	    private static void min_sen_act(){
	  	  
	  	  Vector<umdes.State> expanded_Qc = new Vector<umdes.State>();
	  	  
	  	  //deep copy
	  	  getStateSet().states_unchanged.clear();
	  	  getStateSet().clearEvents();
	  	  for(int k = 0; k<getStateSet().states.size(); k++){
	  		  expanded_Qc.add(getStateSet().states.get(k));
	  		  if(!getStateSet().states_unchanged.contains(getStateSet().states.get(k)))
	  			  getStateSet().states_unchanged.add(getStateSet().states.get(k));
	  	  }

	  	  //for all transitions
	  	  for(int i = 0; i<transOrder.size(); i++){

	  		  if(!automaton.isUnobserv(transOrder.get(i))){

	  			  expanded_Qc = automaton.unObservableReach(expanded_Qc,((RegExpr) transOrder.get(i)));
	  			  if(!violate_Te_spec(expanded_Qc)){
	  				  getStateSet().E_ua.add(((RegExpr) transOrder.get(i))); 
	  				  //add all states after Qc ends (set Qc = expanded_Qc)
	  				  while(getStateSet().states.size()!=expanded_Qc.size()){
	  					  getStateSet().states.add(expanded_Qc.get(getStateSet().states.size()));
	  				  }
	  			  }
	  			  else{
	  				  getStateSet().E_a.add(((RegExpr) transOrder.get(i)));  
	  				  //remove all states after Qc ends (set expanded_Qc = Qc)
	  				  while(getStateSet().states.size()!=expanded_Qc.size()){
	  					  expanded_Qc.remove(getStateSet().states.size());
	  				  }
	  			  }  			  
	  		  }
	  		  else getStateSet().E_uo.add(((RegExpr) transOrder.get(i)));  //it is unobservable

	  	  }
	  	  return;
	    }
	    private static Qc getStateSet(){
	    	return automaton.getStateSet();
	    }
	    
	    private static Vector<StatePair> formatSpec(String inputText) throws Exception{
	    	Vector<StatePair> result = new Vector<StatePair>();
	    	
			String line, fullText = inputText;
			StringTokenizer tokens = new StringTokenizer(fullText, "\n", false);
			while (tokens.hasMoreTokens()) {
				line = tokens.nextToken();
				line = line.trim();
				if(line.equals("")) continue;
				
    			//find where white space is(space,tab)
    			int delimPosition = line.indexOf(' ');
    			if(delimPosition==-1) delimPosition = line.indexOf('\t');
    			if(delimPosition==-1) {
    				throw new Exception("line does not contain two states");
    			}  //there are not two states
    			
    			String str1 = line.substring(0, delimPosition).trim();
    			String str2 = line.substring(delimPosition).trim();
    			
    			if(automaton.hasState(str1)&&automaton.hasState(str2)){
    				result.add(new StatePair(automaton.getState(str1), 
    						                 automaton.getState(str2)));
    			} else throw new Exception("states do not exist");

			}
	    	return result;
	    }
	    
	    /**
	     * drag and drop user interface to enter the transitions in the order that you want to 
	     * consider them when using the online sensor activation algorithms.
	     * @param workDir the working directory of the .transorder file
	     * @throws if the input file is incomplete it will throw a general Exception
	     */
	     private static void orderTransitions(){
	    	  
	        //automaton.getTrans_order().clear();
	   	     	    
	    	 Vector<RegExpr> alphabet = automaton.getAlphabet();
	        
	        for(int i = 0; i<alphabet.size(); i++){
	        	if(automaton.isUnobserv(alphabet.get(i))){
	        		transOrder.add(alphabet.get(i));
	        		continue;	    	
	        	}
	        }
	        Vector<String> events = automaton.getObservableEvents();
	        EventList ev = new EventList(events, new JFrame(), false);
	   	    if(ev.getEventOrder()==null||ev.getEventOrder().size()==0) {System.exit(0);}

	   	    EventOrder.clear();
	   	    for(int i = 0; i<ev.getEventOrder().size(); i++){
	    	   	transOrder.add(new RegExpr((String) ev.getEventOrder().get(i)));
	    	   	EventOrder.add(((String) ev.getEventOrder().get(i)));
	    	   }

	      }
	     /**
	      * prints the expanded t_spec 
	      * @param workDir : path to the output
	      * @throws IOException
	      */
	     private static void printTSpec(BufferedWriter output) throws IOException{
	   	  	  
	    	 Vector<StatePair> v = automaton.getTSpec().violate_vector();
	    	 output.write("t_spec expanded:\n");
	    	 for(int i = 0; i<v.size();i++){
	    		 if((i % 6)==0&&i!=0) output.write("\n");
	    		 if(i+1!=v.size())
	    			 output.write(" { " +  v.get(i).first.name + ", "
	    					 + v.get(i).second.name + "} ,");
	    	 }
	    	 //don't want a comma for the last one
	    	 output.write(" { " +  v.get(v.size()-1).first.name + ", " 
	    			 + v.get(v.size()-1).second.name + "}  \n");
	   		  
	    	 output.write("------------------------------------------------------------");
	     }
	     /**
	      * tests to see if the states in the vector are illegal according to T_spec
	      * @param v The vector of states that is to be tested
	      * @return true if it is illegal, false if it is a good set of states
	      */
	     private static boolean violate_Te_spec(Vector<umdes.State> v){
	   	  
	   	  if(v==null)return false;
	   	  Vector<StatePair> illegal= automaton.getTSpec().violate_vector();
	   	  
	   	  for(int i = 0; i<v.size();i++){
	   		  for(int j = 0; j<v.size(); j++){
	   			  StatePair s = new StatePair((v.get(i)),(v.get(j)));
	   			  if(illegal.contains(s)) return true;
	   		  }
	   	  }	  
	   	  return false;
	     } 

	   //Added by Matthew Hillyard - November 14, 2007
	     /**
	      * Displays a JOptionPane with the given parameters, and waits for the user to pick an option.
	      * Once the user picks an option, this method returns the integer value corresponding to to location in list where that option is found.
	      * 
	      * @param title The title to be displayed on the option pane.
	      * @param description The description of what we want the user to select.
	      * @param list The list of options the user has to choose from.
	      * @param initialSelectionValue The value to be initially selected
	      * 
	      * @return 0 if the user hits the cancel button. Otherwise, the location of the selected item in list. 
	      */
	     private static int getUserInput(String title, String description, String[] list, Object initialSelectionValue)
	     {
	   	  Object selectedValue = JOptionPane.showInputDialog(null, description, title,
	   	  			JOptionPane.INFORMATION_MESSAGE, null, list, initialSelectionValue);
	   	  for(int i = 0; i < list.length; i++)
	   	  {
	   		  if(selectedValue == list[i])
	   		  {
	   			  return i + 1;
	   		  }
	   	  }
	   	  return 0;
	     }
}
