package umdes;

import java.util.Iterator;
import java.util.Vector;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Collections;;

/**
 * 
 * This class contains the set of states an automata can be in, 
 * along with the sets of activated and unactivated sensors, and 
 * the unobservable events.
 * 
 * @author John O'Brien 5/27/09
 *
 */
public class Qc {
	
	 //the states that the automata can be in
	public Vector<State> states; 
	public Vector<State> states_unchanged;
	
	public Set<RegExpr> E_a;   //the activated sensors for the given states
	public Set<RegExpr> E_ua;  //the deactivated sensors for the given states	
	public Set<RegExpr> E_uo;  //the unobservable events
	
	public HashMap<RegExpr,Qc> destStates; // the destination states from E_a
	public boolean isInitial;
	
	public Qc(){
		states = new Vector<State>();
		states_unchanged = new Vector<State>();
		E_a  = new HashSet<RegExpr>();
		E_ua = new HashSet<RegExpr>();
		E_uo = new HashSet<RegExpr>();
		destStates = new HashMap<RegExpr,Qc>();
		isInitial = false;		
	}
	/*public Qc(Qc tmp){
		tmp.states = new Vector(this.states);
		tmp.states_unchanged = new Vector(this.states_unchanged);
		tmp.E_a = new HashSet(this.E_a);
		tmp.E_ua = new HashSet(this.E_ua);
		tmp.E_uo = new HashSet(this.E_uo);
	}*/
	
	//override java.lang.equals
	public boolean equals(Object o){
			
		if(o==null)return false;
		Qc Qc_tmp;
		if (o instanceof Qc) {
			Qc_tmp = (Qc) o;
		}
		else return false;
		if(this.states_unchanged.size()!=Qc_tmp.states_unchanged.size()) return false;
		for(int i = 0; i < this.states_unchanged.size(); i++){
			if(!Qc_tmp.states_unchanged.contains(((State) this.states_unchanged.get(i)))) return false;
		}
		if(this.states.size()!=Qc_tmp.states.size()) return false;
		for(int i = 0; i < this.states.size(); i++){
			if(! Qc_tmp.states.contains(((State) this.states.get(i)))) return false;
		}
		return true;
	}
	//override java.lang.hashcode: Note: there may be better hashcodes, but it needs to be defined
	public int hashCode(){
		int hash = 0;
		for(int i = 0; i < this.states_unchanged.size(); i++){
			hash = hash + this.states_unchanged.get(i).hashCode();
		}
		for(int i = 0; i < this.states.size(); i++){
			hash = hash + this.states.get(i).hashCode();
		}
		return hash;
	}
	
	public void clearAll(){
		states.clear();
		states_unchanged.clear();
		E_a.clear();  
		E_ua.clear();  
		E_uo.clear();		
	}
	public void clearEvents(){
		E_a.clear();  
		E_ua.clear();  
		E_uo.clear();		
	}
	
	public String outputStates(){
		String tmp = "{";
		for(int i = 0; i< states.size()-1; i++){
			tmp = tmp + ((State) states.get(i)).name + ", ";
		}
		if(states.size()!=0)tmp = tmp + ((State) states.get(states.size()-1)).name;
		tmp = tmp + "}";
		return tmp;
	}
	public String outputStatesUnchanged(){
		String tmp = "{";
		for(int i = 0; i< states_unchanged.size()-1; i++){
			tmp = tmp + ((State) states_unchanged.get(i)).name + ", ";
		}
		if(states_unchanged.size()!=0)tmp = tmp + ((State) states_unchanged.get(states_unchanged.size()-1)).name;
		tmp = tmp + "}";
		return tmp;
	}
	public String policyEA(){
		String tmp = "";
	    Iterator<RegExpr> it = E_a.iterator();
  	    if(it.hasNext()) 	    tmp = tmp +"{";
  	    while(it.hasNext()){
  		    tmp = tmp +((RegExpr) it.next()).getExpression();
  	    	if(it.hasNext()) 	    tmp = tmp +", ";
  	    	else tmp = tmp +"} ";
  	    }
  	    return tmp;
	}
	public String policyEUA(){
		String tmp = "";
		Vector<String> sortEvents = new Vector<String>();
	    Iterator<RegExpr> it = E_ua.iterator();
  	    if(it.hasNext()) 	    tmp = tmp +"{";
  	    while(it.hasNext()){
  	    	sortEvents.add(it.next().getExpression());
  	    }
  	    Collections.sort(sortEvents);
  	    for(int i = 0; i < sortEvents.size() -1; i++){
  	    	tmp += ((String) sortEvents.get(i)) + ", ";
  	    }
  	    if(sortEvents.size()>0)
  	    	tmp += ((String) sortEvents.get(sortEvents.size()-1)) + "} ";
  	    
  	    return tmp;
	}
	public String policyEUO(){
		String tmp = "";
	    Iterator<RegExpr> it = E_uo.iterator();
  	    if(it.hasNext()) 	    tmp = tmp +"{";
  	    while(it.hasNext()){
  		    tmp = tmp + it.next().getExpression();
  	    	if(it.hasNext()) 	    tmp = tmp +", ";
  	    	else tmp = tmp +"} ";
  	    }
  	    return tmp;
	}
	

}
