/*** Giddes
 *     - Graphical Interface for Designing Discrete Event Systems
 *  S.Laurie Ricker Ph.D
 *
 *  Developed by I. Fraser & W. Fowler
 *
 *
 * reproduction without the express permission of Toba will
 * result in being taken outside and shot many times with a
 * very tiny gun.
 ***/

//*********************************************************
// Documentation completed January 16, 2002               *
//*********************************************************
// the package
package umdes;


import java.util.Vector;
import java.util.regex.*;
import java.util.Enumeration;
import java.util.Stack;

/**
 * This is the State Object for a Finite Automata. It holds only a name
 *
 * @see UmdesAutomaton
 */
public class State
{
    /**
     * This is the name of a Dead State
     */
    public static final String DEAD= "DEAD";
    // private members
    private static String whiteSpace = "[\\W]*";
    private static String literal = "[\\w_]+";
    static String wsLit = whiteSpace + literal + whiteSpace;
    static String commaLit = wsLit + ",";
    static String literalList = "[" + commaLit + "]*" + wsLit;
    static String compoundLit = "\\{" + literalList  + "\\}";
    static String term = "(" + wsLit + "|" + compoundLit + ")";
    static String commaTerm = term + ",";
    static String termList = "[" + commaTerm + "]*" +  term;
    static String full = "\\(" + termList + "\\)";

    // a name is either our initial literal definition
    // or, it is fully specified.
    static String validName = "(" + full + "|" + wsLit + "|" + compoundLit + ")";


  /**
   * The name of the State
   */
  protected String name;



  //***************************************************************
  // public members
  //***************************************************************
  /**
   * Creates a new State with the name s
   *
   * @param s a String specifiying the name of the new State
   *
   * @throws Exception if the String s is not a valid State name
   */
  public State(String s){
//    if(isValidName(s))
      name= new String(s);
//    else
//      throw new Exception("Invalid State Name");
//
  }

  /**
   * Copy Constructor - usually State.clone() is used
   *
   * @param s a String specifiying the name of the new State
   */
  public State(State s){
    name=s.getName();
  }

  /**
   * Special constructor to be used by StateEstimate to
   * generate a State with the StateEstimate's complex name
   * The usual name checks are not done, and the given string
   * is expected to be correct
   *
   * @param s the new State name
   */
  protected State(StateEstimate s){
    name=new String(s.toString());
  }

  /**
   * This Method will return the name of the state as a String
   * The return is the same as toString(), however it can not
   * be overloaded by a child class
   *
   * @return the name of the State
   *
   */
  public final String getName(){
    return new String(name);
  }
  /**
   * This function sets the name of a State
   *
   * @param s the state's new name
   *
   * @throws Exception if the String s is not a valid State name
   */
  public void setName(String s) throws Exception{
    if(isValidName(s))
      name= new String(s);
    else
      throw new Exception("Invalid State Name");
  }

  /**
   * Returns true if the object s is equal to the State
   *
   * @param s the object to be tested for equality agais this State
   *
   * @return true if the object s is equal to the State
   */
  public boolean equals(Object s){
    if(this==s)  {
      // if it is the same reference (ie: same instance)

      return(true);
    }

    if (s==null){
      // if s is null or not
      //of the same class

      return(false);
    }

    State snew; // check for Class Casting
    try
    {
      snew=(State)s;
    }
    catch (ClassCastException cce)
    {
      return false;
    }

    return(name.equals(snew.getName()));
  }

  /**
   * Returns a unique int hashCode
   *
   * @return the hashCode of this State
   */
  public int hashCode(){
    return(name.hashCode());
  }

  /**
   * Returns a clone of the State
   *
   * @return a clone of the State
   */
  public Object clone(){
    State s;

    try{
      s = new State(this);
    }
    catch(Exception e){
      return(null);
    }

    return(s);
  }
    protected static boolean myValidLiteral(String s) {
	if (s == null)
	    return false;

	if  (Pattern.matches(wsLit,s))
	    return true;
	return false;
    }

    protected static boolean myValidCompoundLiteral(String s) {
	if (s == null)
	    return false;

	if (Pattern.matches(compoundLit,s))
	    return true;
	return false;
    }

    public static boolean myValidName(String s) {
	if (s == null)
	    return false;

	if (Pattern.matches(validName,s))
	    return true;

	return false;
    }


  /**
   * Method checks that string can be a valid State name
   * Invalid characters are: ',' '{' '}' '(' ')'
   *
   * @param s the String to check
   *{1|{1,2,3,5,7,9}|{1,2,4,8}}
   * @return True if the String is a valid State name
   */
  public static boolean isValidName(String s)
  {
	  Vector<Character> allowableCharacters = new Vector<Character>(2); // currently only one allowable character.
      Stack<Character> balancedBrackets = new Stack<Character>();
      Character ch;

      // Make a list of allowable characters here.
      allowableCharacters.add(new Character('_'));
      allowableCharacters.add(new Character('-'));
      
      for(int i=0; i<s.length(); i++)
      {
      	ch = new Character(s.charAt(i));
      	if(ch.charValue() == '{')
      		balancedBrackets.push(ch);
      	else if(ch.charValue() == '}')
      		balancedBrackets.pop();
      	else if(ch.charValue() == ',' && balancedBrackets.empty())
      		return false;
      	//else if(!(Character.isLetterOrDigit(ch.charValue()) || allowableCharacters.contains(ch)))
      		//return false;
      }
      return balancedBrackets.empty();
  	
  }
//  public static boolean isValidName(String s){
//    StringBuffer buffer= new StringBuffer(s);
//
//    char c, lastchar;
//    int i;
//    boolean composite=false;
//    boolean open_found=false;
//    boolean closed_found=false;
//    Character charClass;
//    lastchar='a'; // initialize to start
//
//    for(i=0;i<buffer.length();i++){
//      c=buffer.charAt(i); // grab char
//      switch(c){ // check for reserved chars
//        case '{' :
//
////           if(buffer.length()==1)
////             return false;
////           if(!open_found)
////             open_found=true;
////           //else {
////             //return false;
////          //}
////
////           if(i!=0) // if open brace make sure its the first char and if second in a row
////             return false;
////           else // check for composite state and set flag
////             composite=true;
//           break;
//        case '}' :
////            if(!closed_found)
////              closed_found=true;
//////            else
//////              return false;
////
////            if(i+1!=buffer.length()) // if closing brace make sure its the last char and if second in a row
////              return false;
//            break;
//
//        case ',' :
////            if(lastchar==c||!composite) // check for two comma's in a row
////              return false;
//            break;
//      }
//    }
//
//    return true;
//  }


  /**
   * Parses this State's name into a list of States that combine to
   * name this State.
   * Eg) {2,3,4}  ---> a Vector containing:2 3 4 as separate elements
   *     2        ---> a Vector containing:2
   *
   *
   * @return Vector of composite States
   */
  public Vector<String> getCompositeStates() throws Exception{
    Vector<String> compositeStates = new Vector<String>();
    StringBuffer buffer=new StringBuffer(name);
    int i,start=0;
    String temp;

    char c;

    // check to see if it needs to be parsed
    c=buffer.charAt(0);
    if(c!='{'){ // no parsing needed so add to tempNames
      compositeStates.add(buffer.toString());
      return(compositeStates);
    }

    start=1;
    for(i=2;i<buffer.length();i++){ //start at char 2 since it is a 'well-formed' name
      c=buffer.charAt(i); // grab char

      if(c==','){ // end of substring
        temp=buffer.substring(start,i); // grab substring
        compositeStates.add(temp); // add
        start=i+1;
      }
      else if(c=='}' && i+1==buffer.length()){ // last char
        temp=buffer.substring(start,i);
        compositeStates.add(temp);
      }
    }

    return(compositeStates);
  }

  /**
   * Converts the State into a String.
   *
   * @return String representation of the State
   */
  public String toString(){
    return(name);
  }

  /**
   * Returns true if this State has the name s contained inside it, otherwise
   * false. The name of the State includes the name of its
   * composite States as well as any brackets, braces or commas
   * that are added to display.
   *
   * @param s The String to test
   *
   * @return True if s is the name of the State
   */
  public boolean containsCompositeState(String s) throws Exception{
    return(getCompositeStates().contains(s));
  }

  /**
   * Returns true if the name of the State is equal to the given
   * String (given by String.equals(s)).
   *
   * @param s the String to test
   *
   * @return True if s is equal to the name, else false
   */
  public boolean hasName(String s){
    return(name.equals(s));
  }

  /**
   * This method takes two States and combines them into one.
   * The two given States become composite States of the new State
   * If an error occurs null is returned
   *
   * @param s1 the first State to combine
   * @param s2 the second State to combine
   *
   * @return the combined State or if an error occurs a null object
   */
  public static State combineStates(State state1,State state2) throws Exception{
    if(state1.equals(state2)) return(new State(state1));

    Vector<String> state1Names = state1.getCompositeStates(); // parse s1
    Vector<String> state2Names = state2.getCompositeStates(); // parse s2

    // combine
    Enumeration<String> e = state2Names.elements();
    while(e.hasMoreElements()){
      String tempString = e.nextElement();
      
      for(int i = 0;i < state1Names.size(); i++){
    	  if(tempString.compareTo(state1Names.elementAt(i))>0) continue;
    	  else{
    		  if(!state1Names.contains(tempString))
    			  state1Names.insertElementAt(tempString,i);
    		  break;
    	  }
      }
    }
    // create the new State name (with opening brace);
    String tempString = new String("{");
    e = state1Names.elements();
    while(e.hasMoreElements()){
      tempString += e.nextElement();
      if(e.hasMoreElements()) // if not at the end of the list
        tempString+=",";
    }
    // add closing brace
    tempString+="}";

    return (new State(tempString));
  }
  

  public boolean containsCompositeState(State s) throws Exception{
    Vector<String> names=getCompositeStates();
    return(names.contains(s.toString()));
  }


};




