package umdes;

import java.util.Vector;

public class T_Conf {

	//private boolean reverse = false;
	private Vector<Window> sensor_on;
	private Vector<Window> sensor_off;
	private Vector<Window> window_order;
	private Vector<Window> unObservableWindows;
	
	private Vector<Window> last_off;
	private Vector<WindowPair> last_winPair;
	
	private Vector<Window> D = new Vector<Window>();
	
	protected Vector<WindowPair> t_conf; //vector of WindowPair's
	
	T_Conf(UmdesAutomaton machine, Vector<Window> sensor_on_in, Vector<Window> windowPartitions_in){
		t_conf = new Vector<WindowPair>();
		unObservableWindows = getUnobservableWindows(machine);
		last_off = new Vector<Window>();
		sensor_on = new Vector<Window>(sensor_on_in);
		sensor_off = new Vector<Window>();
		last_winPair = new Vector<WindowPair>();
		window_order = windowPartitions_in;
		
		
		//start with T = {(x,x): x belongs to X}
		Vector<Window> win_states = new Vector<Window>();
		for(int i = 0; i < machine.getStates().size(); i++){
			win_states = getWindows((State) machine.getStates().get(i), machine);
			for(int count1 = 0; count1 < win_states.size(); count1++){
				for(int count2 = 0; count2 < win_states.size(); count2++){
					Window w1 = (Window) win_states.get(count1);
					Window w2 = (Window) win_states.get(count2);
					WindowPair w_pair = new WindowPair(w1, w2);	
					if(!t_conf.contains(w_pair)) t_conf.add(w_pair);
				}
			}
		}
		win_states = null;
		
		expandTConf(machine);
		last_winPair.clear();

			
	}
	/**
	 * this will expand T_conf to include the unobservable reach and the strings 
	 * that are needed to keep it feasible. This is equivalent to equations (28)
	 * and (29) from the paper this is based on. (Optimal Sensor Activation for
	 * Diagnosing Discrete Event Systems)
	 * @param machine
	 */
	public void expandTConf(UmdesAutomaton machine){
		Vector<Transition> trans1 = new Vector<Transition>();
		Vector<Transition> trans2 = new Vector<Transition>();
		Vector<Window> window_new1 = new Vector<Window>();
		Vector<Window> window_new2 = new Vector<Window>();

		//add unobservable reach
		for(int i = 0; i < t_conf.size(); i++){
			State s1 = t_conf.get(i).w1.state;
			State s2 = t_conf.get(i).w2.state;

			try{
				trans1 = machine.getEvents().getTransitions(s1);
			} catch (Exception e){}
			try{
				trans2 = machine.getEvents().getTransitions(s2);
			} catch (Exception e){}
			
			for(int j = 0; trans1!=null &&j < trans1.size(); j++){
				for(int k = 0; trans2!=null && k < trans2.size(); k++){
					window_new1.clear();
					window_new2.clear();
					RegExpr r1 = trans1.get(j).getLabel().getEvent();
					RegExpr r2 = trans2.get(k).getLabel().getEvent();
					
					if(machine.isUnobserv(r1)||isSensorOff(s1,r1)) {
						window_new1 		= getWindows(trans1.get(j).to, machine);
						Vector<Window> curWindowVec = getWindows(trans1.get(j).from, machine);
						addWindowPair( window_new1, curWindowVec, null);
					}
					if(machine.isUnobserv(r2)||isSensorOff(s1,r2)) {
						window_new2 = getWindows(trans2.get(k).to, machine);	
						Vector<Window> curWindowVec = getWindows(trans1.get(j).from, machine);						
						addWindowPair(curWindowVec, window_new2, null);
					}

					
					if(r1.equals(r2)){
						window_new1 = getWindows(trans1.get(j).to, machine);
						window_new2 = getWindows(trans2.get(k).to, machine);
					}
											
					if(window_new1.isEmpty()&&window_new2.isEmpty()) continue;


					if(window_new1.isEmpty()){ //window1 stays stationary and window2 moves
						addWindowPair(window_new1, window_new2, t_conf.get(i).w1);
						continue;
					}
					if(window_new2.isEmpty()){ //window2 stays stationary and window1 moves
						addWindowPair(window_new1, window_new2, t_conf.get(i).w2);
						continue;
					}				
					addWindowPair(window_new1, window_new2, null);
					
				}
			}
		}
		removeInfeasible(machine);
	}
	
	/**
	 * this will turn off the sensors that make the machine infeasible and re-expand the TConf 
	 * This is equivalent to equations (30)from the paper this is based on. 
	 * (Optimal Sensor Activation for Diagnosing Discrete Event Systems)
	 */
	public void removeInfeasible(UmdesAutomaton machine){
		
		//because t_conf is reflexive, the checking the first elements is equivalent to 
		//checking the first and second elements
		for(int i = 0; i < t_conf.size(); i++){
			
			Window w1 = ((WindowPair) t_conf.get(i)).w1;
			Window w2 = ((WindowPair) t_conf.get(i)).w2;
			
			//if the events are equal and the states are not
			if(w1.e.equals(w2.e)&&!w1.state.equals(w2.state)){
				//and one of the sensors is off and one is on
				if(sensor_off.contains(w1)&&!sensor_off.contains(w2)){ 
					turnOffWindow(w2);
					expandTConf(machine); //recurse to expand the t_conf
				}
			}
			
		}
		
		
	}
	/**
	 * adds checks to see if all the WindowPair formed by the two vectors are in t_conf 
	 * if they are not there it adds both the pair and the reflexive pair
	 * @param w1 A vector of windows
	 * @param w2 A vector of windows
	 * @param win If one of the vectors is empty it will use this as the other window
	 *  	in adding the pair
	 */
	private void addWindowPair(Vector<Window> winVec1, Vector<Window> winVec2, Window win){
		
		Window window1, window2;
		WindowPair w_pair_new;
		
		if(winVec1.isEmpty()&&winVec2.isEmpty()) return;
		
		if(winVec1.isEmpty()){
			for(int i = 0; i < winVec2.size(); i++){
				window2 = winVec2.get(i);
				w_pair_new = new WindowPair(win, window2);
				if(!t_conf.contains(w_pair_new)){
					t_conf.add(w_pair_new);	
					last_winPair.add(w_pair_new);
					w_pair_new = new WindowPair(window2, win); //add the reflexive pair
					if(!t_conf.contains(w_pair_new)){
						t_conf.add(w_pair_new);	
						last_winPair.add(w_pair_new);
					}
				}
			}
			return;
		}
		if(winVec2.isEmpty()){
			for(int i = 0; i < winVec1.size(); i++){
				window1 = (Window) winVec1.get(i);
				w_pair_new = new WindowPair(win, window1);
				if(!t_conf.contains(w_pair_new)){
					t_conf.add(w_pair_new);	
					last_winPair.add(w_pair_new);
					w_pair_new = new WindowPair(window1, win); //add the reflexive pair
					if(!t_conf.contains(w_pair_new)){
						last_winPair.add(w_pair_new);
						t_conf.add(w_pair_new);	
					}
				}
			}		
			return;
		}
		for(int i = 0; i < winVec1.size(); i++){
			for(int j = 0; j < winVec2.size(); j++){

				window1 = (Window) winVec1.get(i);
				window2 = (Window) winVec2.get(j);
				w_pair_new = new WindowPair(window1, window2);
				if(!t_conf.contains(w_pair_new)){
					t_conf.add(w_pair_new);	
					last_winPair.add(w_pair_new);
					w_pair_new = new WindowPair(window2, window1); //add the reflexive pair
					if(!t_conf.contains(w_pair_new)){
						last_winPair.add(w_pair_new);
						t_conf.add(w_pair_new);	
					}

				}			
			}
		}
		
		
	}
	public boolean isWindowOff(Window window){
		return isSensorOff(window.state, window.e);
	}
	private boolean isSensorOff(State s, RegExpr r){
		
		if(sensor_off==null || sensor_off.size()==0) return false;
		for(int i = 0; i < sensor_off.size(); i++){
			Window w = (Window) sensor_off.get(i);
			if(w.state.equals(s)&&w.e.equals(r)) return true;
		}		
		return false;
	}
	
	/**
	 * gets the sensor activation policy and converts it to a two dimensional vector with transitions 
	 * so it is in the format the cluster diagnoser can use, this may need to be amended when window 
	 * size is greater than 1.
	 * @param : machine
	 * @return : a Vector of Vector's containing Transitions
	 */
	public Vector<Vector<Transition>> getPolicy(UmdesAutomaton machine){
		
		Vector<Vector<Transition>> policy = new Vector<Vector<Transition>>();
		Vector<Transition> innerVector = new Vector<Transition>();
		
		for(int i = 0; i < sensor_off.size(); i++){		
			Vector<Transition> trans = machine.getEvents().getTransition(sensor_off.get(i).state, sensor_off.get(i).e); 
			if(trans==null) continue;
			for(int j = 0; j < trans.size(); j++){
				innerVector.add(trans.get(j));
			}			
		}
		policy.add(innerVector);
		
		return policy;
	}
	/**
	 * gets all the windows with the state s
	 * @param s: the state
	 * @return: a Vector of all of the desire windows
	 */
	private Vector<Window> getWindows(State s, UmdesAutomaton machine){
		Vector<Window> v = new Vector<Window>();
		for(int i = 0; i < window_order.size(); i++){
			if(((Window) window_order.get(i)).state.equals(s)) v.add(window_order.get(i));
		}
		for(int i = 0; i < unObservableWindows.size(); i++){
			if(((Window) unObservableWindows.get(i)).state.equals(s)) v.add(unObservableWindows.get(i));
		}
		return v;
	}
	
	private Vector<Window> getUnobservableWindows(UmdesAutomaton machine){
		Vector<Window> v = new Vector<Window>();
		Vector<Transition> trans;
		for(int i = 0; i < machine.getStates().size(); i++){
			try{
				trans = machine.getEvents().getTransitions(machine.getStates().get(i));
			} catch (Exception e){ continue;}
			for(int j = 0; j < trans.size(); j++){
				if(machine.isUnobserv(trans.get(j).getLabel().getEvent()))
					v.add(new Window(machine.getStates().get(i), 1, new Vector<String>(), trans.get(j).getLabel().getEvent()));
			}
		}
		return v;
	}
	public int winSize(){
		return window_order.size();
	}
	public Window getWindow(int i){
		if(window_order.isEmpty()|| i < 0 || i > window_order.size() )return null;
		return (Window) window_order.get(i);
	}
	public int sensorOffSize(){
		return sensor_off.size();
	}	
	public Window getSensorOff(int i){
		if(sensor_off.isEmpty()|| i < 0 || i > sensor_off.size() )return null;
		return (Window) sensor_off.get(i);
	}
	
	public void turnOffWindow(Window w){
		last_off.add(w);		
		if(!sensor_off.contains(w)) sensor_off.add(w);
		sensor_on.remove(w);
	}
	public void resetSensors(){
		sensor_off.removeAll(last_off);
		sensor_on.addAll(last_off);
		last_off.clear();
		t_conf.removeAll(last_winPair);
		last_winPair.clear();
	}
	public void keepSensorChanges(){	
		last_off.clear();
		last_winPair.clear();
	}
	public boolean isInD(Window w){
		return D.contains(w);
	}
	public void addToD(Window w){
		if(!D.contains(w)) D.add(w);
	}
}
