package umdes;

import java.util.Vector;
import java.lang.Math;

/**
 * 
 * this is the class used to hold the set of cluster automatas for diagnosis
 * @author John O'Brien
 *
 */
public class Xd implements Cloneable{


	private Cluster curPsuedo;		  	//the current state of the Psuedo automata
	private Vector<Cluster> startState; 	//Vector of the initial clusters of all clusterAutomatas
	private Vector<Cluster>  curState;    //Vector of the current clusters of all clusterAutomatas
	private Vector<ClusterAutomata>  automatas;  //Vector of ClusterAutomatas
	private ClusterAutomata Psuedo;   	//the Psuedo cluster automatas
	private Vector<Transition> trans_out;
	
	
	public Xd(Vector<ClusterAutomata> automatas_in){
		
		setTrans_out(new Vector<Transition>());	
		startState = new Vector<Cluster>();
		setCurState(new Vector<Cluster>());
		setAutomatas(automatas_in);
		
		for(int i = 0; i < getAutomatas().size(); i++){
			startState.add(((ClusterAutomata) getAutomatas().get(i)).initial);
		}
		setCurState(startState.clone());  //may need to do deeper copy??
		setPsuedo(findPsuedo(getAutomatas())); 
		curPsuedo = getPsuedo().initial;
		
	}
	
	private Xd(Vector<Cluster> auto_clusters, Xd cur){
		setTrans_out(new Vector<Transition>());
		setCurState(auto_clusters);
		setAutomatas(cur.getAutomatas());
		startState = cur.startState;
		setPsuedo(cur.getPsuedo());
		for(int i = 0; i < getCurState().size(); i++){
			if(((Cluster) getCurState().get(i)).id==0)  curPsuedo = (Cluster) getCurState().get(i);
		}
		
	}
	public Cluster getCurPsuedo(){
		Cluster c;
		try{
			c = (Cluster) curPsuedo.clone();
		} catch (CloneNotSupportedException e) {c = null;}
		return c;
	}
	
	/**
	 * finds the Psuedo automata in a Vector of ClusterAutomatas
	 * @param clust_auto
	 * @return
	 */
	private ClusterAutomata findPsuedo(Vector<ClusterAutomata> clust_auto){
			
		for(int i = 0; i < clust_auto.size(); i++){
			if(clust_auto.get(i).isPsuedo) return clust_auto.get(i);
		}
		return null;
		
	}

	
	
	public Vector<Xd> getXdOut(Vector<Vector<Cluster>> output_states, RegExpr r){
		//output_states is the clusters that should be the clusterAutomata's curCluster
		
		Vector<Xd> output = new Vector<Xd>();
		int numOutput = 1;
		for(int i = 0; i < output_states.size(); i++){
			if(output_states.get(i).size() > 1) numOutput = numOutput * (int) Math.pow(2,output_states.get(i).size()-1);
		}
		int[] curPositionInVec = new int[output_states.size()];
		for(int i = 0; i < curPositionInVec.length; i++){ 
			if(output_states.get(i).size() == 0 || output_states.get(i).size() == 1) curPositionInVec[i] = -1; 
			else curPositionInVec[i] = 0; } //initialize to 0
		
		//boolean getNextPosition = true;
		
		for(int numCalculated = 0; numCalculated < numOutput; numCalculated++){
			
			Vector<Cluster> automatas_in = new Vector<Cluster>();		
			for(int i = 0; i < output_states.size();  i++ ){

			
				if(output_states.get(i).size() == 0) automatas_in.add(getCurState().get(i));
				
				if(output_states.get(i).size() == 1) automatas_in.add(output_states.get(i).get(0));
				
				if(output_states.get(i).size() > 1){
					automatas_in.add(((Vector<Cluster>) output_states.get(i)).get(curPositionInVec[i]));		
				}
						
			}
			curPositionInVec = updatePos(output_states, curPositionInVec);
			Xd tmp_xd = new Xd(automatas_in, this);
			if(!output.contains(tmp_xd)){
				Transition t = new Transition(new State(this.toString()), new State(tmp_xd.toString()) ,r);
				if(!getTrans_out().contains(t)) getTrans_out().add(t);
				output.add(tmp_xd);
			}
		}
		
		return output;
	}
	
	private int[] updatePos(Vector<Vector<Cluster>> output_states, int[] initPos){
		
		boolean carrying = false;
		
		for(int i = initPos.length - 1; i >= 0; i--){
			
			//only one element and doesn't need to updated
			if(initPos[i]== -1) continue;
			
			//simple case in which there is no carry bits
			if(!(initPos[i] + 1 >  output_states.get(i).size() -1) && !carrying){
				initPos[i]++;
				return initPos;
			}
			if((initPos[i] + 1 >  output_states.get(i).size() -1)){
				carrying = true;
				continue;
			}
			
			if(!(initPos[i] + 1 >  output_states.get(i).size() -1) && carrying){	
				initPos[i]++;
				if(i + 1 > initPos.length) return initPos;
				for(int j = i+1; j < initPos.length; j++){
					if(initPos[j]== -1) continue;
					else initPos[j] = 0;
				}
				return initPos;
			}
					
		}
		
		
		return initPos;
	}
	
	
	//make clone method
	@SuppressWarnings("unchecked")
	protected Object clone() throws CloneNotSupportedException {

	    Xd clone = (Xd)super.clone();
	    
	    clone.setCurState((Vector<Cluster>)this.getCurState().clone());
	    clone.curPsuedo = (Cluster)this.curPsuedo.clone();
	    
	    return clone;

	  }
	
	public boolean equals(Object o){
		
		if(o==null)return false;
		Xd c_tmp;
		if (o instanceof Xd) {
			c_tmp = (Xd) o;
		}
		else return false;	
		if(!this.getPsuedo().equals(c_tmp.getPsuedo())) return false;
		if(!this.startState.equals(c_tmp.startState)) return false;
		if(!this.getCurState().equals(c_tmp.getCurState())) return false;
		if(!this.getAutomatas().equals(c_tmp.getAutomatas())) return false;
		
		return true;
	
	}

	public int hashCode(){
		int hash = 0;
		hash = getPsuedo().hashCode()+ startState.hashCode() + getCurState().hashCode() + getAutomatas().hashCode();
		return hash;
	}
	
	public String toString() {
		String str  = "[";
		for(int i = 0; i < getCurState().size(); i++){
			Cluster c = (Cluster) getCurState().get(i);
			if(c.id!=0) str += "C" + c.id + "(" + c.startState + "),"; 
			else{
				str += "(Cp(" + c.startState + "),"; 
				if(c.Nfk==true) str += "Nf)";
				else str += "f)";
			}
		}
		str += "]";
		
		return str;
	}

	public void setAutomatas(Vector<ClusterAutomata> automatas) {
		this.automatas = automatas;
	}
	public Vector<ClusterAutomata> getAutomatas() {
		return automatas;
	}
	public void setCurState(Vector<Cluster> curState) {
		this.curState = curState;
	}
	
	@SuppressWarnings("unchecked")
	public void setCurState(Object curState) {
		if(curState instanceof Vector){
			if(curState!=null){				
				if(((Vector<?>) curState).get(0) instanceof Cluster)
					this.curState = (Vector<Cluster>) curState;
			}
		}
		
	}
	public Vector<Cluster> getCurState() {
		return curState;
	}
	public void setPsuedo(ClusterAutomata psuedo) {
		Psuedo = psuedo;
	}
	public ClusterAutomata getPsuedo() {
		return Psuedo;
	}
	public void setTrans_out(Vector<Transition> trans_out) {
		this.trans_out = trans_out;
	}
	public Vector<Transition> getTrans_out() {
		return trans_out;
	}
}
